# The MIT License (MIT)
# Copyright © 2022 Opentensor Foundation

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated 
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation 
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, 
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of 
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
# DEALINGS IN THE SOFTWARE.

from argparse import ArgumentParser

from munch import Munch
from ruamel.yaml import YAML

yaml = YAML()

class Config ( Munch ):
    """
    Implementation of the config class
    """
    def __init__(self, loaded_config = None ):
        super().__init__()
        if loaded_config:
            raise NotImplementedError('Function load_from_relative_path is not fully implemented.')

    def __repr__(self) -> str:
        return self.__str__()
    
    def __str__(self) -> str:
        return "\n" + yaml.dump(self.toDict())

    def to_string(self, items) -> str:
        """ Get string from items
        """
        return "\n" + yaml.dump(items.toDict())

    def update_with_kwargs( self, kwargs ):
        """ Add config to self
        """
        for key,val in kwargs.items():
            self[key] = val

class InvalidConfigFile(Exception):
        """ In place of YAMLError
        """

def config_builder( parser: ArgumentParser = None, strict: bool = False ) -> Config:
        r""" Translates the passed parser into a nested config.
        Args:
            parser (argparse.Parser):
                Command line parser object.
            strict (bool):
                If true, the command line arguments are strictly parsed.
        Returns:
            config (runpodpy.Config):
                Nested config object created from parser arguments.
        """
        if parser == None:
            parser = ArgumentParser()
       
        params = parser.parse_known_args()[0]
        
        _config = Config()

        # Splits params on dot syntax i.e machine.podName          
        for arg_key, arg_val in params.__dict__.items():
            split_keys = arg_key.split('.')
            head = _config
            keys = split_keys
            while len(keys) > 1:
                if hasattr(head, keys[0]):
                    head = getattr(head, keys[0])  
                    keys = keys[1:]   
                else:
                    head[keys[0]] = Config()
                    head = head[keys[0]] 
                    keys = keys[1:]
            if len(keys) == 1:
                head[keys[0]] = arg_val

        return _config



# The MIT License (MIT)
# Copyright © 2022 Opentensor Foundation

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import argparse
import asyncio
import sys

from gql.transport.aiohttp import AIOHTTPTransport
from loguru import logger
from ruamel.yaml import YAML
import runpodpy

from runpodpy.cli import create, destroy, list_pods, start, stop
from runpodpy.runpod import RunPod
from runpodpy.config import Config, config_builder
from runpodpy import CloudType, GPUTypeId

logger = logger.opt(colors=True)


def configure_logging(config):
    logger.remove()
    if config.debug == True:
        logger.add(sys.stderr, level="TRACE")
    else:
        logger.add(sys.stderr, level="INFO")


async def run_command(config: Config, command):
    """Main function"""
    global RUNPOD_API_KEY
    if config.runpod_api.get('API_KEY') is None:
        raise ValueError('No API_KEY found in config.runpod_api')
        
    RUNPOD_API_KEY = config.runpod_api["API_KEY"]

    # Connect to runpod api
    runpod_transport = AIOHTTPTransport(
        url=f"{config.runpod_api['URL']}?api_key={RUNPOD_API_KEY}",
        headers={
            "content-type": "application/json",
        },
    )

    # Setup runpod API
    runpod = RunPod(runpod_transport)

    if await runpod.test_connection(logger):
        logger.success("Connected to runpod api")

        await command(runpod, config, logger)
    else:
        logger.error("Failed to connect to runpod")
        exit(1)

def main():
    parser = argparse.ArgumentParser(
        description=f"RunPodPy v{runpodpy.__version__}",
        add_help=True,
    )
    parser._positionals.title = "commands"
    command_parsers = parser.add_subparsers(dest="command")
    parser.add_argument(
        "--debug", action="store_true", help="Enable debug logging", default=False
    )
    parser.add_argument(
        "--config_file",
        type=str,
        help="Path to config file",
        default=None,
        required=False,
    )
    parser.add_argument(
        "--runpod_api.URL",
        "--URL",
        type=str,
        dest="runpod_api.URL",
        help="Base URL of runpod api",
        default="https://api.runpod.io/graphql",
        required=False,
    )
    parser.add_argument(
        "--runpod_api.API_KEY",
        "--API_KEY",
        type=str,
        dest="runpod_api.API_KEY",
        help="Your RunPod.io api key",
        required=False,
    )

    stop_parser = command_parsers.add_parser("stop", help="Stop a pod")
    stop_parser.add_argument(
        "--podId",
        type=str,
        help="Pod id to stop",
        required=False,
        default=None,
    )
    stop_parser.add_argument(
        "--all",
        action="store_true",
        help="Stop all pods",
        required=False,
        default=False,
    )

    destroy_parser = command_parsers.add_parser(
        "destroy", help="Destroy a pod", aliases=["terminate"]
    )
    destroy_parser.add_argument(
        "--podId",
        type=str,
        help="Pod id to destroy",
        required=False,
        default=None,
    )
    destroy_parser.add_argument(
        "--all",
        action="store_true",
        help="Destroy all pods",
        required=False,
        default=False,
    )

    start_parser = command_parsers.add_parser(
        "start", help="Start a pod", aliases=["run", "resume"]
    )
    start_parser.add_argument(
        "--podId",
        type=str,
        help="Pod id to start",
        required=False,
        default=None,
    )
    start_parser.add_argument(
        "--all",
        action="store_true",
        help="Start all pods",
        required=False,
        default=False,
    )
    start_parser.add_argument(
        "--max_bid",
        type=float,
        help="Maximum bid for the pod",
        required=False,
        default=None,
    )
    start_parser.add_argument(
        "--spot",
        action="store_true",
        help="Start the pod as a spot instance",
        required=False,
        default=False,
    )

    create_parser = command_parsers.add_parser("create", help="Create a pod")
    create_parser.add_argument(
        "--spot",
        action="store_true",
        help="Create the pod as a spot instance",
        required=False,
        default=False,
    )
    create_parser.add_argument(
        "--max_bid",
        type=float,
        help="Maximum bid for the pod",
        required=False,
        default=None,
    )    
    create_parser.add_argument(
        "--podName",
        "--machine.podName",
        dest="machine.podName",
        type=str,
        help="Pod name to create",
        required=False,
        default=None,
    )
    create_parser.add_argument(
        "--gpuTypeId",
        "--machine.gpuTypeId",
        dest="machine.gpuTypeId",
        type=GPUTypeId,
        choices=list(GPUTypeId),
        help='The GPU type to use. e.g. "NVIDIA GeForce RTX 3080 Ti"',
        required=False,
        default=GPUTypeId.RTX_3080_TI,
    )
    create_parser.add_argument(
        "--cloudType",
        "--cloud",
        dest="cloudType",
        type=CloudType,
        choices=list(CloudType),
        help='The CloudType to deploy to. e.g. COMMUNITY, SECURE',
        required=False,
        default=CloudType.COMMUNITY,
    )
    create_parser.add_argument(
        "--imageName",
        "--machine.imageName",
        dest="machine.imageName",
        type=str,
        help="The docker image to use",
        required=False,
        default='pytorch/pytorch',
    )
    create_parser.add_argument(
        "--volumePath",
        "--machine.volumePath",
        dest="machine.volumePath",
        type=str,
        help="The volume path to mount",
        required=False,
        default='/root',
    )
    create_parser.add_argument(
        "--args",
        "--machine.args",
        dest="machine.args",
        type=str,
        help='The arguments to pass to docker. e.g. "bash -c "sleep infinity""',
        required=False,
        default='bash -c "sleep infinity"',
    )
    create_parser.add_argument(
        "--containerDiskSize",
        "--machine.containerDiskSize",
        dest="machine.containerDiskSize",
        type=int,
        help="The size of the container disk (GB)",
        required=False,
        default=40,
    )
    create_parser.add_argument(
        "--volumeSize",
        "--machine.volumeSize",
        dest="machine.volumeSize",
        type=int,
        help="The size of the volume (GB)",
        required=False,
        default=40,
    )
    create_parser.add_argument(
        "--gpuCount",
        "--machine.gpuCount",
        dest="machine.gpuCount",
        type=int,
        help="The number of GPUs to use",
        required=False,
        default=1,
    )
    create_parser.add_argument(
        "--templateId",
        "--machine.templateId",
        dest="machine.templateId",
        type=str,
        help="The templateId to use",
        required=False,
        default=argparse.SUPPRESS,
    )

    list_parser = command_parsers.add_parser("list", help="List pods")

    config: Config = config_builder(parser)

    yaml = YAML()

    if config.config_file:
        # Load config file into config
        with open(config.config_file, "r") as config_file:
            config.update(yaml.load(config_file))

    commands = {
        "stop": stop,
        "destroy": destroy,
        "terminate": destroy,
        "start": start,
        "run": start,
        "resume": start,
        "create": create,
        "list": list_pods,
    }

    configure_logging(config)

    if config.command in commands.keys():
        # Run main with the command
        asyncio.run(run_command(config, commands[config.command]))
    else:
        parser.print_help()
        exit(1)


if __name__ == "__main__":
    
    main()



# The MIT License (MIT)
# Copyright © 2022 Opentensor Foundation

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

__version__ = "0.3.2"

from enum import Enum


class CloudType(Enum):
    COMMUNITY = "COMMUNITY"
    SECURE = "SECURE"

    def __str__(self):
        return self.value


class GPUTypeId(Enum):
    RTX_A4000 = "NVIDIA RTX A4000"
    RTX_A4500 = "NVIDIA RTX A4500"
    RTX_A5000 = "NVIDIA RTX A5000"
    RTX_A6000 = "NVIDIA RTX A6000"
    A100_80GB = "NVIDIA A100 80GB PCIe"
    A40 = "NVIDIA A40"
    RTX_3070 = "NVIDIA GeForce RTX 3070"
    RTX_3080 = "NVIDIA GeForce RTX 3080"
    RTX_3080_TI = "NVIDIA GeForce RTX 3080 Ti"
    RTX_3090 = "NVIDIA GeForce RTX 3090"
    V100_FHHL = "Tesla V100-FHHL-16GB"
    V100_SXM2 = "Tesla V100-SXM2-16GB"
    TESLA_V100 = "Tesla V100-PCIE-16GB"

    def __str__(self):
        return self.value

    @classmethod
    def from_gpuDisplayName(cls, s: str) -> "GPUTypeId":
        s_ = s.replace(" ", "_").upper()
        typeId = getattr(GPUTypeId, s_, None)
        if typeId is None:
            raise ValueError(f"Unknown GPU type: {s}")
        return typeId

class PodStatus(Enum):
    """
    Status (desired status) of a pod.
    """
    RUNNING = "RUNNING"
    EXITED = "EXITED"

    def __str__(self):
        return self.value



# The MIT License (MIT)
# Copyright © 2022 Opentensor Foundation

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from __future__ import annotations

import asyncio
import json
from typing import Dict, List, Optional

import loguru
from gql import Client, gql
from gql.transport.aiohttp import AIOHTTPTransport
from gql.transport.exceptions import TransportQueryError, TransportServerError

from runpodpy import CloudType, GPUTypeId, PodStatus


class RunPodException(Exception):
    """RunPodException"""

    pass


class OutbidException(RunPodException):
    """OutbidException
    Raised when the bid for a spot instance is higher than the current bid.
    """

    pass


class RunPodInstance:
    """RunPodInstance"""

    ip_address: str = "ssh.runpod.io"
    gpuCount: int
    gpuDisplayName: GPUTypeId
    vcpuCount: int
    memoryInGb: int
    imageName: str
    podName: str
    cost: float
    podId: str
    podHostId: str
    spot: bool
    desiredStatus: PodStatus
    cloudType: CloudType

    def __init__(
        self,
        podName: str,
        cost: float,
        podId: str,
        podHostId: str,
        imageName: str,
        gpuDisplayName: GPUTypeId,
        gpuCount: int,
        vcpuCount: int,
        memoryInGb: int,
        cloudType: CloudType,
        desiredStatus: PodStatus,
        spot: bool = True,
    ):
        self.podName = podName
        self.cost = cost
        self.spot = spot
        self.podId = podId
        self.podHostId = podHostId
        self.gpuCount = gpuCount
        self.gpuDisplayName = gpuDisplayName
        self.vcpuCount = vcpuCount
        self.memoryInGb = memoryInGb
        self.cloudType = cloudType
        self.imageName = imageName
        self.desiredStatus = desiredStatus

    def __str__(self):
        return f"{self.podId} - {self.gpuDisplayName} - ${self.cost:0.2f}/hr - {self.cloudType.value}:{'SPOT' if self.spot else 'ON-DEMAND'}:{self.desiredStatus.value}"

    async def stop_instance(self, runpod: "RunPod", logger: loguru.Logger) -> bool:
        """Stops the instance"""
        await runpod.stop_instance(self.podId, logger)

    async def destroy_instance(self, runpod: "RunPod", logger: loguru.Logger) -> bool:
        """Destroys the instance"""
        await runpod.destroy_instance(self.podId, logger)

    async def start_instance(
        self, runpod: "RunPod", logger: loguru.Logger, max_bid: Optional[float] = None
    ) -> bool:
        """Starts the instance"""
        await runpod.start_instance(
            self.podId, self.gpuCount, logger, max_bid, self.spot
        )


class RunPod:
    """RunPod class"""

    gql_transport: AIOHTTPTransport = None

    def __init__(self, gql_transport: AIOHTTPTransport):
        self.gql_transport = gql_transport

    async def __create_spot_instance(
        self,
        max_bid: float,
        podName: str,
        imageName: str,
        containerDiskSize: int,
        volumeSize: int,
        volumePath: str,
        gpuCount: int,
        gpuTypeId: GPUTypeId,
        minVcpuCount: int,
        minMemoryInGb: int,
        args: str,
        logger: loguru.Logger,
        cloudType: CloudType = CloudType.COMMUNITY,
    ) -> RunPodInstance:
        """Creates a new spot instance using the API"""

        try:
            async with Client(
                transport=self.gql_transport,
                fetch_schema_from_transport=False,
            ) as session:
                params = {
                    "podName": podName,
                    "imageName": imageName,
                    "containerDiskSize": containerDiskSize,
                    "volumeSize": volumeSize,
                    "volumePath": volumePath,
                    "gpuCount": gpuCount,
                    "gpuTypeId": gpuTypeId.value,
                    "args": args,
                    "max_bid": max_bid,
                    "minVcpuCount": minVcpuCount,
                    "minMemoryInGb": minMemoryInGb,
                    "cloudType": cloudType.value,
                }

                query = gql(
                    """mutation {{
                                podRentInterruptable(
                                    input: {{
                                        bidPerGpu: {max_bid},
                                        cloudType: {cloudType},
                                        gpuCount: {gpuCount},
                                        volumeInGb: {volumeSize},
                                        containerDiskInGb: {containerDiskSize},
                                        minVcpuCount: {minVcpuCount},
                                        minMemoryInGb: {minMemoryInGb},
                                        gpuTypeId: "{gpuTypeId}",
                                        name: "{podName}",
                                        
                                        dockerArgs: {args},
                                        volumeMountPath: "{volumePath}",
                                        imageName: "{imageName}",

                                        templateId: "{templateId}",
                                    }}
                                ) {{
                                    id
                                    imageName
                                    podType
                                    machineId
                                    costPerHr
                                    gpuCount
                                    vcpuCount
                                    memoryInGb
                                    desiredStatus
                                    machine {{
                                        podHostId
                                        gpuDisplayName
                                    }}
                                }}
                            }}""".format(
                        **params
                    )
                )

                data = await session.execute(query)
                pod_data: Dict[str, str] = data["podRentInterruptable"]

                pod: RunPodInstance = RunPodInstance(
                    podName=params["podName"],
                    cost=float(pod_data["costPerHr"]),
                    podId=pod_data["id"],
                    podHostId=pod_data["machine"]["podHostId"],
                    spot=pod_data["podType"] == "INTERRUPTABLE",
                    gpuDisplayName=gpuTypeId,
                    gpuCount=int(pod_data["gpuCount"]),
                    vcpuCount=int(pod_data["vcpuCount"]),
                    memoryInGb=int(pod_data["memoryInGb"]),
                    cloudType=cloudType,
                    imageName=imageName,
                    desiredStatus=PodStatus[pod_data["desiredStatus"]],
                )

                return pod

        except (TransportServerError, TransportQueryError) as e:
            if "outbid" in e.errors[0]["message"]:
                raise OutbidException(e.errors[0]["message"])
            else:
                logger.exception(e)
            return None

    async def __create_spot_instance_from_template_id(
        self,
        max_bid: float,
        podName: str,
        templateId: str,
        containerDiskSize: int,
        volumeSize: int,
        gpuCount: int,
        gpuTypeId: GPUTypeId,
        minVcpuCount: int,
        minMemoryInGb: int,
        logger: loguru.Logger,
        cloudType: CloudType = CloudType.COMMUNITY,
    ) -> RunPodInstance:
        """Creates a new spot instance using the API"""

        try:
            async with Client(
                transport=self.gql_transport,
                fetch_schema_from_transport=False,
            ) as session:
                params = {
                    "podName": podName,
                    "templateId": templateId,
                    "containerDiskSize": containerDiskSize,
                    "volumeSize": volumeSize,
                    "gpuCount": gpuCount,
                    "gpuTypeId": str(gpuTypeId),
                    "minVcpuCount": minVcpuCount,
                    "minMemoryInGb": minMemoryInGb,
                    "max_bid": max_bid,
                    "cloudType": str(cloudType)
                }

                query = gql(
                    """mutation {{
                                podRentInterruptable(
                                    input: {{
                                        bidPerGpu: {max_bid},
                                        cloudType: {cloudType},
                                        gpuCount: {gpuCount},
                                        volumeInGb: {volumeSize},
                                        containerDiskInGb: {containerDiskSize},
                                        minVcpuCount: {minVcpuCount},
                                        minMemoryInGb: {minMemoryInGb},
                                        gpuTypeId: "{gpuTypeId}",
                                        name: "{podName}",
                                        templateId: "{templateId}",
                                    }}
                                ) {{
                                    id
                                    imageName
                                    podType
                                    machineId
                                    costPerHr
                                    gpuCount
                                    vcpuCount
                                    memoryInGb
                                    imageName
                                    desiredStatus
                                    machine {{
                                        podHostId
                                        gpuDisplayName
                                    }}
                                }}
                            }}""".format(
                        **params
                    )
                )

                data = await session.execute(query)
                pod_data: Dict[str, str] = data["podRentInterruptable"]

                pod: RunPodInstance = RunPodInstance(
                    podName=params["podName"],
                    cost=float(pod_data["costPerHr"]),
                    podId=pod_data["id"],
                    podHostId=pod_data["machine"]["podHostId"],
                    spot=pod_data["podType"] == "INTERRUPTABLE",
                    gpuDisplayName=gpuTypeId,
                    gpuCount=int(pod_data["gpuCount"]),
                    vcpuCount=int(pod_data["vcpuCount"]),
                    memoryInGb=int(pod_data["memoryInGb"]),
                    cloudType=cloudType,
                    imageName=pod_data["imageName"],
                    desiredStatus=PodStatus[pod_data["desiredStatus"]],
                )

                return pod

        except (TransportServerError, TransportQueryError) as e:
            if "outbid" in e.errors[0]["message"]:
                raise OutbidException(e.errors[0]["message"])
            else:
                logger.exception(e)
            return None

    async def __create_on_demand_instance(
        self,
        podName: str,
        imageName: str,
        containerDiskSize: int,
        volumeSize: int,
        volumePath: str,
        gpuCount: int,
        gpuTypeId: GPUTypeId,
        minVcpuCount: int,
        minMemoryInGb: int,
        args: str,
        logger: loguru.Logger,
        cloudType: CloudType = CloudType.COMMUNITY,
    ) -> RunPodInstance:
        """Creates a new spot instance using the API"""
        try:
            async with Client(
                transport=self.gql_transport,
                fetch_schema_from_transport=False,
            ) as session:
                # make post request with bearer token
                params = {
                    "podName": podName,
                    "imageName": imageName,
                    "containerDiskSize": containerDiskSize,
                    "volumeSize": volumeSize,
                    "volumePath": volumePath,
                    "gpuCount": gpuCount,
                    "gpuTypeId": gpuTypeId.value,
                    "args": args,
                    "minVcpuCount": minVcpuCount,
                    "minMemoryInGb": minMemoryInGb,
                    "cloudType": cloudType.value,
                }

                query = gql(
                    """mutation {{
                                podFindAndDeployOnDemand(
                                    input: {{
                                        cloudType: {cloudType},
                                        gpuCount: {gpuCount},
                                        volumeInGb: {volumeSize},
                                        containerDiskInGb: {containerDiskSize},
                                        minVcpuCount: {minVcpuCount},
                                        minMemoryInGb: {minMemoryInGb},
                                        gpuTypeId: "{gpuTypeId}",
                                        name: "{podName}",
                                        imageName: "{imageName}",
                                        dockerArgs: {args},
                                        volumeMountPath: "{volumePath}",
                                    }}
                                ) {{
                                    id
                                    imageName
                                    podType
                                    machineId
                                    costPerHr
                                    gpuCount
                                    vcpuCount
                                    memoryInGb
                                    desiredStatus
                                    machine {{
                                        podHostId
                                        gpuDisplayName
                                    }}
                                }}
                            }}""".format(
                        **params
                    )
                )

                data = await session.execute(query)
                print(data)
                pod_data: Dict[str, str] = data["podRentInterruptable"]

                pod: RunPodInstance = RunPodInstance(
                    podName=params["podName"],
                    cost=float(pod_data["costPerHr"]),
                    podId=pod_data["id"],
                    podHostId=pod_data["machine"]["podHostId"],
                    spot=pod_data["podType"] == "INTERRUPTABLE",
                    gpuDisplayName=gpuTypeId,
                    gpuCount=int(pod_data["gpuCount"]),
                    vcpuCount=int(pod_data["vcpuCount"]),
                    memoryInGb=int(pod_data["memoryInGb"]),
                    cloudType=cloudType,
                    imageName=imageName,
                    desiredStatus=PodStatus[pod_data["desiredStatus"]]
                )

                return pod

        except (TransportServerError, TransportQueryError) as e:
            if "outbid" in e.errors[0]["message"]:
                raise OutbidException(e.errors[0]["message"])
            else:
                logger.exception(e)
            return None

    async def __create_on_demand_instance_from_template_id(
        self,
        podName: str,
        templateId: str,
        containerDiskSize: int,
        volumeSize: int,
        gpuCount: int,
        gpuTypeId: GPUTypeId,
        minVcpuCount: int,
        minMemoryInGb: int,
        logger: loguru.Logger,
        cloudType: CloudType = CloudType.COMMUNITY,
    ) -> RunPodInstance:
        """Creates a new spot instance using the API"""
        try:
            async with Client(
                transport=self.gql_transport,
                fetch_schema_from_transport=False,
            ) as session:
                # make post request with bearer token
                params = {
                    "podName": podName,
                    "templateId": templateId,
                    "containerDiskSize": containerDiskSize,
                    "volumeSize": volumeSize,
                    "gpuCount": gpuCount,
                    "gpuTypeId": gpuTypeId.value,
                    "minVcpuCount": minVcpuCount,
                    "minMemoryInGb": minMemoryInGb,
                    "cloudType": cloudType.value,
                }

                query = gql(
                    """mutation {{
                                podFindAndDeployOnDemand(
                                    input: {{
                                        cloudType: {cloudType},
                                        gpuCount: {gpuCount},
                                        volumeInGb: {volumeSize},
                                        containerDiskInGb: {containerDiskSize},
                                        minVcpuCount: {minVcpuCount},
                                        minMemoryInGb: {minMemoryInGb},
                                        gpuTypeId: "{gpuTypeId}",
                                        name: "{podName}",
                                        templateId: "{templateId}",
                                    }}
                                ) {{
                                    id
                                    imageName
                                    podType
                                    machineId
                                    costPerHr
                                    gpuCount
                                    vcpuCount
                                    memoryInGb
                                    imageName
                                    desiredStatus
                                    machine {{
                                        podHostId
                                        gpuDisplayName
                                    }}
                                }}
                            }}""".format(
                        **params
                    )
                )

                data = await session.execute(query)
                print(data)
                pod_data: Dict[str, str] = data["podRentInterruptable"]

                pod: RunPodInstance = RunPodInstance(
                    podName=params["podName"],
                    cost=float(pod_data["costPerHr"]),
                    podId=pod_data["id"],
                    podHostId=pod_data["machine"]["podHostId"],
                    spot=pod_data["podType"] == "INTERRUPTABLE",
                    gpuDisplayName=gpuTypeId,
                    gpuCount=int(pod_data["gpuCount"]),
                    vcpuCount=int(pod_data["vcpuCount"]),
                    memoryInGb=int(pod_data["memoryInGb"]),
                    cloudType=cloudType,
                    imageName=pod_data["imageName"],
                    desiredStatus=PodStatus[pod_data["desiredStatus"]],
                )

                return pod

        except (TransportServerError, TransportQueryError) as e:
            if "outbid" in e.errors[0]["message"]:
                raise OutbidException(e.errors[0]["message"])
            else:
                logger.exception(e)
            return None

    async def create_instance(
        self,
        max_bid: float,
        podName: str,
        imageName: str,
        containerDiskSize: int,
        volumeSize: int,
        volumePath: str,
        gpuCount: int,
        gpuTypeId: GPUTypeId,
        minVcpuCount: int,
        minMemoryInGb: int,
        args: str,
        logger: loguru.Logger,
        cloudType: CloudType = CloudType.COMMUNITY,
        spot: bool = True,
    ) -> RunPodInstance:
        """Creates a new instance"""
        pod: RunPodInstance = None
        args = json.dumps(args)
        if spot:
            pod = await self.__create_spot_instance(
                max_bid,
                podName,
                imageName,
                containerDiskSize,
                volumeSize,
                volumePath,
                gpuCount,
                gpuTypeId,
                minVcpuCount,
                minMemoryInGb,
                args,
                logger,
                cloudType,
            )
        else:
            pod = await self.__create_on_demand_instance(
                podName,
                imageName,
                containerDiskSize,
                volumeSize,
                volumePath,
                gpuCount,
                gpuTypeId,
                minVcpuCount,
                minMemoryInGb,
                args,
                logger,
                cloudType,
            )

        if pod is None:
            return None

        podId = pod.podId

        pod_ = await self.get_pod_by_id(podId, logger)
        while pod_ is None:
            await asyncio.sleep(3)  # wait for pod to finish start
            pod_ = await self.get_pod_by_id(podId, logger)

        return pod

    async def create_instance_from_template_id(
        self,
        max_bid: float,
        podName: str,
        templateId: str,
        containerDiskSize: int,
        volumeSize: int,
        gpuCount: int,
        gpuTypeId: GPUTypeId,
        minVcpuCount: int,
        minMemoryInGb: int,
        logger: loguru.Logger,
        cloudType: CloudType = CloudType.COMMUNITY,
        spot: bool = True,
    ) -> RunPodInstance:
        """Creates a new instance"""
        pod: RunPodInstance = None
        if spot:
            pod = await self.__create_spot_instance_from_template_id(
                max_bid,
                podName,
                templateId,
                containerDiskSize,
                volumeSize,
                gpuCount,
                gpuTypeId,
                minVcpuCount,
                minMemoryInGb,
                logger,
                cloudType,
            )
        else:
            pod = await self.__create_on_demand_instance_from_template_id(
                podName,
                templateId,
                containerDiskSize,
                volumeSize,
                gpuCount,
                gpuTypeId,
                minVcpuCount,
                minMemoryInGb,
                logger,
                cloudType,
            )

        if pod is None:
            return None

        podId = pod.podId

        pod_ = await self.get_pod_by_id(podId, logger)
        while pod_ is None:
            await asyncio.sleep(3)  # wait for pod to finish start
            pod_ = await self.get_pod_by_id(podId, logger)

        return pod

    async def get_pod_by_id(
        self, podId: str, logger: loguru.Logger
    ) -> Optional[RunPodInstance]:
        """Gets a pod from the API by its podId"""
        try:
            async with Client(
                transport=self.gql_transport,
                fetch_schema_from_transport=False,
            ) as session:

                params = {
                    "podId": podId,
                }

                query = gql(
                    """query myPods($podId: String!) {
                            pod (input: {podId: $podId}) {
                                id
                                podType
                                costPerHr
                                name
                                gpuCount
                                vcpuCount
                                memoryInGb
                                imageName
                                desiredStatus
                                machine {
                                    podHostId
                                    gpuDisplayName
                                    secureCloud
                                }
                            }
                            
                    }"""
                )

                data = await session.execute(query, params)

                pod_data: Dict[str, str] = data["pod"]

                podName = pod_data["name"]
                podId = pod_data["id"]
                podHostId = pod_data["machine"]["podHostId"]
                cost = float(pod_data["costPerHr"]) if "costPerHr" in pod_data else None
                # Not a typo, enum has a spelling issue in API
                spot = pod_data["podType"] == "INTERRUPTABLE"
                try:
                    gpuDisplayName = GPUTypeId.from_gpuDisplayName(
                        pod_data["machine"]["gpuDisplayName"]
                    )
                except ValueError:
                    gpuDisplayName = None
                gpuCount = int(pod_data["gpuCount"])
                vcpuCount = int(pod_data["vcpuCount"])
                memoryInGb = int(pod_data["memoryInGb"])
                imageName = pod_data["imageName"]
                secureCloud: bool = pod_data["machine"]["secureCloud"]

                pod = RunPodInstance(
                    podName=podName,
                    cost=cost,
                    podId=podId,
                    podHostId=podHostId,
                    spot=spot,
                    gpuDisplayName=gpuDisplayName,
                    gpuCount=gpuCount,
                    vcpuCount=vcpuCount,
                    memoryInGb=memoryInGb,
                    imageName=imageName,
                    cloudType=CloudType.SECURE if secureCloud else CloudType.COMMUNITY,
                    desiredStatus=PodStatus[pod_data["desiredStatus"]]
                )

                return pod
        except TransportQueryError as e:
            if (
                len(e.errors) >= 1
                and e.errors[0]["extensions"]["code"] == "INTERNAL_SERVER_ERROR"
            ):
                # podId not found
                return None
            else:
                logger.exception(e)
                logger.exception(query)
                raise RunPodException(f"Failed to get pod {podId} with error: {e}")

        except TransportServerError as e:
            logger.exception(e)
            logger.exception(query)
            raise RunPodException(f"Failed to get pod {podId} with error: {e}")

    async def get_pods(self, logger: loguru.Logger) -> List[RunPodInstance]:
        """Gets all the pods from the API"""
        try:
            async with Client(
                transport=self.gql_transport,
                fetch_schema_from_transport=False,
            ) as session:
                query = gql(
                    """query myPods {
                            myself {
                                pods {
                                    id
                                    name
                                    podType
                                    gpuCount
                                    vcpuCount
                                    memoryInGb
                                    imageName
                                    costPerHr
                                    desiredStatus
                                    machine {
                                        podHostId
                                        gpuDisplayName
                                        secureCloud
                                    }
                                }
                            }
                    }"""
                )

                data = await session.execute(query)
                pods_data: List[Dict[str, str]] = data["myself"]["pods"]
                pods: List[RunPodInstance] = []

                for pod_data in pods_data:
                    podName = pod_data["name"]
                    podId = pod_data["id"]
                    podHostId = pod_data["machine"]["podHostId"]
                    cost = (
                        float(pod_data["costPerHr"])
                        if "costPerHr" in pod_data
                        else None
                    )
                    # Not a typo, enum has a spelling issue in API
                    spot = pod_data["podType"] == "INTERRUPTABLE"
                    try:
                        gpuDisplayName = GPUTypeId.from_gpuDisplayName(
                            pod_data["machine"]["gpuDisplayName"]
                        )
                    except ValueError:
                        gpuDisplayName = None
                    gpuCount = int(pod_data["gpuCount"])
                    vcpuCount = int(pod_data["vcpuCount"])
                    memoryInGb = int(pod_data["memoryInGb"])
                    imageName = pod_data["imageName"]
                    secureCloud: bool = pod_data["machine"]["secureCloud"]

                    pod = RunPodInstance(
                        podName=podName,
                        cost=cost,
                        podId=podId,
                        podHostId=podHostId,
                        spot=spot,
                        gpuDisplayName=gpuDisplayName,
                        gpuCount=gpuCount,
                        vcpuCount=vcpuCount,
                        memoryInGb=memoryInGb,
                        imageName=imageName,
                        cloudType=CloudType.SECURE if secureCloud else CloudType.COMMUNITY,
                        desiredStatus=PodStatus[pod_data["desiredStatus"]],
                    )
                    pods.append(pod)

                return pods

        except (TransportServerError, TransportQueryError) as e:
            logger.exception(e)
            raise RunPodException(f"Failed to get pods with error: {e}")

    async def test_connection(self, logger: loguru.Logger) -> bool:
        """Tests the connection to the runpod api"""
        try:
            async with Client(
                transport=self.gql_transport,
                fetch_schema_from_transport=False,
            ) as session:
                # make post request with bearer token
                query = gql(
                    """query myPods {
                            myself {
                                id
                            }
                    }"""
                )

                result = await session.execute(query)
                if result is None:
                    return False

                # Authentication is successful if we are not guest
                return result["myself"]["id"] != "guest"

        except (TransportServerError, TransportQueryError) as e:
            logger.exception(e)
            return False

    async def stop_instance(self, podId: str, logger: loguru.Logger) -> bool:
        """Stops an instance by the podId"""
        try:
            async with Client(
                transport=self.gql_transport,
                fetch_schema_from_transport=False,
            ) as session:

                params = {
                    "podId": podId,
                }

                query = gql(
                    """mutation($podId: String!) {
                        podStop(input: {podId: $podId}) {
                            id
                            desiredStatus
                        }
                    }"""
                )

                data = await session.execute(query, params)

                return (
                    data["podStop"]["id"] == podId
                    and data["podStop"]["desiredStatus"] == "EXITED"
                )

        except (TransportServerError, TransportQueryError) as e:
            logger.exception(e)
            logger.exception(query)
            return False

    async def destroy_instance(self, podId: str, logger: loguru.Logger) -> bool:
        """Destroys the instance by the podId"""
        try:
            async with Client(
                transport=self.gql_transport,
                fetch_schema_from_transport=False,
            ) as session:

                params = {
                    "input": {
                        "podId": podId,
                    }
                }

                query = gql(
                    """mutation terminatePod($input: PodTerminateInput!) {
                        podTerminate(input: $input)
                    }"""
                )

                data = await session.execute(query, params)

                return data["podTerminate"] is None

        except (TransportServerError, TransportQueryError) as e:
            logger.exception(e)
            logger.exception(query)
            return False

    async def __start_spot_instance(
        self, podId: str, gpuCount: int, max_bid: float, logger: loguru.Logger
    ) -> bool:
        """Starts a spot instance by the podId with a gpuCount and a max_bid per GPU"""
        try:
            async with Client(
                transport=self.gql_transport,
                fetch_schema_from_transport=False,
            ) as session:

                params = {
                    "podId": podId,
                    "maxBid": max_bid,
                    "gpuCount": gpuCount,
                }

                query = gql(
                    """mutation($podId: String!, $maxBid: Float!) {
                        podBidResume(input: {podId: $podId, bidPerGpu: $maxBid, gpuCount: 1}) {
                            id
                            desiredStatus
                            imageName
                            env
                            machineId
                            machine {
                            podHostId
                            }
                        }
                    }"""
                )

                data = await session.execute(query, params)
                pod_data: Dict[str, str] = data["podBidResume"]

                return (
                    pod_data["id"] == podId and pod_data["desiredStatus"] == "RUNNING"
                )
        except TransportQueryError as e:
            if (
                len(e.errors) >= 1
                and "Cannot resume a pod that is not in exited state"
                in e.errors[0]["message"]
            ):
                logger.info(f"Pod {podId} is already running")
                return True
            else:
                logger.exception(e)
                logger.exception(query)
                return False

        except TransportServerError as e:
            logger.exception(e)
            logger.exception(query)
            return False

    async def __start_on_demand_instance(
        self, podId: str, gpuCount: int, logger: loguru.Logger
    ) -> bool:
        """Starts an on-demand instance by the podId with a gpuCount"""
        try:
            async with Client(
                transport=self.gql_transport,
                fetch_schema_from_transport=False,
            ) as session:

                params = {
                    "podId": podId,
                    "gpuCount": gpuCount,
                }

                query = gql(
                    """mutation($podId: String!, $maxBid: Float!) {
                        podResume(input: {podId: $podId, gpuCount: 1}) {
                            id
                            desiredStatus
                            imageName
                            env
                            machineId
                            machine {
                            podHostId
                            }
                        }
                    }"""
                )

                data = await session.execute(query, params)
                pod_data: Dict[str, str] = data["podResume"]

                return (
                    pod_data["id"] == podId and pod_data["desiredStatus"] == "RUNNING"
                )
        except TransportQueryError as e:
            if (
                len(e.errors) >= 1
                and "Cannot resume a pod that is not in exited state"
                in e.errors[0]["message"]
            ):
                logger.info(f"Pod {podId} is already running")
                return True
            else:
                logger.exception(e)
                logger.exception(query)
                return False

        except TransportServerError as e:
            logger.exception(e)
            logger.exception(query)
            return False

    async def start_instance(
        self,
        podId: str,
        gpuCount: int,
        logger: loguru.Logger,
        max_bid: Optional[float] = None,
        spot: bool = False,
    ):
        """Starts an instance by the podId"""
        if spot:
            return await self.__start_spot_instance(podId, gpuCount, max_bid, logger)
        else:
            return await self.__start_on_demand_instance(podId, gpuCount, logger)



# The MIT License (MIT)
# Copyright © 2022 Opentensor Foundation

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from __future__ import annotations

from typing import List

import loguru
from munch import Munch
from tabulate import tabulate
from runpodpy import CloudType, GPUTypeId

from runpodpy.runpod import OutbidException, RunPod, RunPodException, RunPodInstance


async def stop(runpod: RunPod, config: Munch, logger: loguru.Logger) -> None:
    """Stops the instance(s)"""
    # Check runpod API connection
    if not await runpod.test_connection(logger):
        raise RunPodException("Runpod API is not connected")

    if config.get("podId") is not None:
        # Stop the pod
        await runpod.stop_instance(config.podId, logger)
        logger.info(f"Stopped pod {config.podId}")

    elif config.all:
        logger.info("Stopping all pods")
        # Stop all pods
        pods = await runpod.get_pods(logger)
        for pod in pods:
            await runpod.stop_instance(pod.podId, logger)
            logger.info(f"Stopped pod {pod.podId}")
        logger.info("DONE | Stopped all pods")
    else:
        logger.error("No --podId or --all specified")


async def start(runpod: RunPod, config: Munch, logger: loguru.Logger) -> None:
    """Starts the instance(s)"""
    # Check runpod API connection
    if not await runpod.test_connection(logger):
        raise RunPodException("Runpod API is not connected")

    spot: bool = config.spot

    if config.get("podId") is not None:
        # Get the pod
        pod: RunPodInstance = await runpod.get_pod_by_id(config.podId, logger)
        if pod is None:
            logger.error(f"Pod {config.podId} not found")
            return False
        # Start the pod
        await runpod.start_instance(
            pod.podId, pod.gpuCount, logger, config.max_bid, spot
        )
        logger.info(f"Started pod {config.podId}")

    elif config.all:
        logger.info("Starting all pods")
        # Get all pods
        pods = await runpod.get_pods()
        for pod in pods:
            # Start the pod
            await runpod.start_instance(
                pod.podId, pod.gpuCount, logger, config.max_bid, spot
            )
            logger.info(f"Started pod {pod.podId}")
        logger.info("DONE | Started all pods")
    else:
        logger.error("No --podId or --all specified")


async def create(runpod: RunPod, config: Munch, logger: loguru.Logger) -> None:
    """Creates a runpod instance"""
    # Check runpod API connection
    if not await runpod.test_connection(logger):
        raise RunPodException("Runpod API is not connected")

    if config.machine.get("podName") is None:
        # Get all the pods
        pods = await runpod.get_pods(logger)

        # Get name for pod
        podName: str = config["machine"]["gpuTypeId"].replace(" ", "_") + str(len(pods))

        config.machine["podName"] = podName
    
    config.machine["gpuTypeId"] = GPUTypeId(config.machine["gpuTypeId"])
    config["cloudType"] = CloudType[config["cloudType"].upper()]

    if config.get("max_bid") is None:
        logger.exception("No --max_bid specified")
        return
    try:
        # Create a pod if there are no pods
        if config.machine.get("templateId") is None:
            pod = await runpod.create_instance(
                config.max_bid,
                config.machine["podName"],
                config.machine["imageName"],
                config.machine["containerDiskSize"],
                config.machine["volumeSize"],
                config.machine["volumePath"],
                config.machine["gpuCount"],
                config.machine["gpuTypeId"],
                config.machine["minVcpuCount"],
                config.machine["minMemoryInGb"],
                logger,
                cloudType=config.cloudType,
                spot=config.spot,
            )
        else:
            # Create a pod from a template
            pod = await runpod.create_instance_from_template_id(
                config.max_bid,
                config.machine.get("podName"),
                config.machine.get("templateId"),
                config.machine.get("containerDiskSize"),
                config.machine.get("volumeSize"),
                config.machine.get("gpuCount"),
                config.machine.get("gpuTypeId"),
                config.machine.get("minVcpuCount"),
                config.machine.get("minMemoryInGb"),
                logger,
                cloudType=config.cloudType,
                spot=config.spot,
            )
    except OutbidException as e:
        logger.error(e)
        pod = None
        
    if pod is None:
        logger.error(
            f"Failed to create {config.cloudType} pod - max_bid:{config.max_bid} spot:{config.spot}"
        )
    else:
        logger.info(f"Created {config.cloudType} pod {pod.podId}")


async def destroy(runpod: RunPod, config: Munch, logger: loguru.Logger) -> None:
    """Destroys the instance(s)"""
    # Check runpod API connection
    if not await runpod.test_connection(logger):
        raise RunPodException("Runpod API is not connected")

    if config.get("podId") is not None:
        # Destroy the pod
        await runpod.destroy_instance(config.podId, logger)
        logger.info(f"Destroyed pod {config.podId}")

    elif config.all:
        logger.info("Destroying all pods")
        # Destroy all pods
        pods = await runpod.get_pods(logger)
        for pod in pods:
            await runpod.destroy_instance(pod.podId, logger)
            logger.info(f"Destroyed pod {pod.podId}")
        logger.info("DONE | Destroyed all pods")
    else:
        logger.error("No --podId or --all specified")


async def list_pods(runpod: RunPod, config: Munch, logger: loguru.Logger) -> None:
    """Lists the instance(s)"""
    # Check runpod API connection
    if not await runpod.test_connection(logger):
        raise RunPodException("Runpod API is not connected")

    logger.info("Connected to RunPod API")

    pods = await runpod.get_pods(logger)

    if len(pods) == 0:
        logger.info("No pods found")
    else:
        table: List[List[str]] = [
            [
                "podId",
                "podName",
                "cloudType",
                "instanceType",
                "gpuTypeId",
                "gpuCount",
                "costPerHr",
                "ip_address",
            ]
        ]
        table.extend(
            [
                [
                    pod.podId,
                    pod.podName,
                    pod.cloudType,
                    "SPOT" if pod.spot else "ON_DEMAND",
                    pod.gpuDisplayName,
                    pod.gpuCount,
                    f"${pod.cost:.2f}/hr",
                    pod.ip_address,
                ]
                for pod in pods
            ]
        )
        logger.info(f"Found {len(pods)} pods:\n{tabulate(table, headers='firstrow')}")
